import sys
import os

MOTU_CLIENT_CONFIG_PATH = os.path.join("motu_utils", "cfg")
LINUX_CONFIG_PATH = os.path.join("/usr", "local", MOTU_CLIENT_CONFIG_PATH)
WINDOWS_CONFIG_PATH = os.path.join(sys.prefix, MOTU_CLIENT_CONFIG_PATH)
VIT_ENV_CONFIG_PATH = MOTU_CLIENT_CONFIG_PATH
LOCAL_CONFIG_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "motu_utils", "cfg")

def getConfigPath():
    if os.path.isdir(LINUX_CONFIG_PATH):
    	return LINUX_CONFIG_PATH
    elif os.path.isdir(WINDOWS_CONFIG_PATH):
    	return WINDOWS_CONFIG_PATH
    elif os.path.isdir(LOCAL_CONFIG_PATH):
        return LOCAL_CONFIG_PATH
    elif os.path.isdir(VIT_ENV_CONFIG_PATH):
    	return VIT_ENV_CONFIG_PATH
    else:
    	return ""


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

import sys
if sys.version_info > (3, 0):
    from html.parser import HTMLParser
else:
    from HTMLParser import HTMLParser



class FounderParser(HTMLParser):
    """
    Parser witch found the form/action section an return it
    """
    def __init__(self, *args, **kargs):
        HTMLParser.__init__(self, *args, **kargs)
        self.action_ = None

    def handle_starttag(self, tag, attrs):
        d = dict(attrs)
        if tag == 'form' and 'action' in d:
            self.action_ = d['action']



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Deprecated
# Only for retrocompatibility of Motu <= v1.7
import logging
import motuclient

deprecatedWarnMsg='"motu-client" module is deprecated since version1.8. A new Python module named "motuclient" without a dash separator between motu and client shall be used instead of "motu-client".'
motuclient.initLogger()
motuclient.log.warn( deprecatedWarnMsg )    


def get_client_version():
    #motuclient.log.warn( deprecatedWarnMsg + 'e.g. motuclient.get_client_version()')    
    return motuclient.get_client_version()

def get_client_artefact():
    #motuclient.log.warn( deprecatedWarnMsg + ', e.g. motuclient.get_client_artefact()')  
    return motuclient.get_client_artefact()

def main():
    #motuclient.log.warn( deprecatedWarnMsg + ', e.g. motuclient.main()')  
    motuclient.main()

if __name__ == '__main__':
    #motuclient.log.warn( deprecatedWarnMsg + ', e.g. motuclient.main()')  
    motuclient.main()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
import io, sys, logging

def copy(sourceHandler, destHandler, callback = None, blockSize = 65535 ):
    """Copy the available content through the given handler to another one. Process
    can be monitored with the (optional) callback function.
    
    sourceHandler: the handler through witch downloading content
    destHandler: the handler into which writing data
    callback: the callback function called for each block read. Signature: f: sizeRead -> void
    blockSize: the size of the block used to read data
    
    returns the total size read
    """
    read = 0
    while 1:
        block = sourceHandler.read(blockSize)
        if(isinstance(block, bytes)):
           exit_condition = b''
        else:
           exit_condition = ''
        if block == exit_condition:
           break
        read += len(block)
        try:
          if type(destHandler) == io.StringIO:
            if sys.version_info > (3, 0):
              destHandler.write( str(block, 'utf-8') )
            else:
              destHandler.write( unicode(block, 'utf-8') )
          else:
            destHandler.write(block)
        except Exception as inst:
          log = logging.getLogger("utils_stream:copy")
          log.error("Exception while copying remote data")
          log.error(" - Type = %s", type(inst))    # the exception instance
          if hasattr(inst, 'args'):
            log.error(" - Attributes = %s", inst.args)     # arguments stored in .args
          log.error(" - Full exception = %s", inst) 
          
        callback(read)

    return read


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

import logging
import re
import sys
import time


if sys.version_info > (3, 0):
    import urllib.request, urllib.error
    from urllib.parse import parse_qs, urlparse, quote, quote_plus
else:
    from urllib import quote_plus, quote
    from urlparse import urlparse, parse_qs

from motu_utils import utils_http, utils_log, utils_html, utils_messages
from motu_utils import utils_collection


# pattern used to search for a CAS url within a response
CAS_URL_PATTERN = '(.*)/login.*'

def authenticate_CAS_for_URL(url, user, pwd, **url_config):
    """Performs a CAS authentication for the given URL service and returns
    the service url with the obtained credential.
    
    The following algorithm is done:
    1) A connection is opened on the given URL
    2) We check that the response is an HTTP redirection
    3) Redirected URL contains the CAS address
    4) We ask for a ticket for the given user and password
    5) We ask for a service ticket for the given service
    6) Then we return a new url with the ticket attached
    
    url: the url of the service to invoke
    user: the username
    pwd: the password"""
    
    log = logging.getLogger("utils_cas:authenticate_CAS_for_URL")
    
    server, sep, options = url.partition( '?' )
    
    log.info( 'Authenticating user %s for service %s' % (user,server) )      
    
    delays = [10, 30, 100, 300]
    nbDelays = 4
    tries = 0
    redirected = False
    while not redirected and tries<=nbDelays:
        try:
            connexion = utils_http.open_url(url, **url_config)
            # connexion response code must be a redirection, else, there's an error (user can't be already connected since no cookie or ticket was sent)
            if connexion.url != url:
                # find the cas url from the redirected url
                redirected_url = connexion.url
                p = parse_qs(urlparse(connexion.url).query, keep_blank_values=False)
                redirectServiceUrl = p['service'][0]
                
                
                m = re.search(CAS_URL_PATTERN, redirected_url)
                
                if not m is None:
                    redirected = True
        except Exception as e:
            pass
        if not redirected and tries<nbDelays:
            log.warn("Warning: CAS connection failed, retrying in " + str(delays[tries]) + " seconds ...")
            time.sleep(delays[tries])
            tries = tries + 1
    if not redirected:
        if redirected_url is None:
            raise Exception(
                utils_messages.get_external_messages()['motuclient.exception.authentication.not-redirected'] % server)
        else:
            raise Exception(
                utils_messages.get_external_messages()['motuclient.exception.authentication.unfound-url'] % redirected_url)
    
    url_cas = m.group(1) + '/v1/tickets'

    opts = utils_http.encode(utils_collection.ListMultimap(username = user, password = pwd))

    utils_log.log_url(log, "login user into CAS:\t", url_cas + '?' + opts)
    url_config['data']=opts.encode()
    
    connected = False
    tries = 0
    while not connected and tries<=nbDelays:
        try:
            connexion = utils_http.open_url(url_cas, **url_config)
            connected = True
        except Exception as e:
            if tries<nbDelays:
                log.warn("Warning: Authentication failed, retrying in " + str(delays[tries]) + " seconds ...")
                time.sleep(delays[tries])
                tries = tries + 1
            else:
                if hasattr(e, 'code') and e.code == 400:
                    log.error( """Error: Bad user login or password:
                    
                         On *nix OS, you must use the single quote, otherwise it may expand specific characters.
                         [...] -u 'string' or --user 'string' [...]
                         
                         On Windows OS, you must use the double quote, because single quotes are treated literally.
                         [...] -p "string" or --pwd "string" [...]
                         """)
                raise e
        
    fp = utils_html.FounderParser()
    for line in connexion:
        log.log(utils_log.TRACE_LEVEL, 'utils_html.FounderParser() line: %s', line)
        # py3 compatibility
        if (isinstance(line, bytes)):
            fp.feed(line.decode())
        else:
            fp.feed(line)
        
    tgt = fp.action_[fp.action_.rfind('/') + 1:]
    log.log(utils_log.TRACE_LEVEL, 'TGT: %s', tgt)

    # WARNING : don't use 'fp.action_' as url : it seems protocol is always http never https 
    # use 'url_cas', extract TGT from 'fp.action_' , then construct url_ticket.
    # url_ticket = fp.action_
    url_ticket = url_cas + '/' + tgt

    if url_ticket is None:
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.authentication.tgt'])
    
    utils_log.log_url(log, "found url ticket:\t", url_ticket)

    opts = utils_http.encode(utils_collection.ListMultimap(service = quote_plus(redirectServiceUrl)))

    utils_log.log_url(log, 'Granting user for service\t', url_ticket + '?' + opts)
    url_config['data']=opts.encode()

    validated = False
    tries = 0
    while not validated and tries<=nbDelays:
        try:
            ticket = utils_http.open_url(url_ticket, **url_config).readline()
            validated = True
        except Exception as e:
            if tries<nbDelays:
                log.warn("Warning: Ticket validation failed, retrying in " + str(delays[tries]) + " seconds ...")
                time.sleep(delays[tries])
                tries = tries + 1
            else:
                raise

    # py3 compatibility
    if (isinstance(ticket, bytes)):
        ticket = ticket.decode()

    utils_log.log_url(log, "found service ticket:\t", ticket)
    
    # we append the download url with the ticket and return the result  
    service_url = redirectServiceUrl + '&ticket=' + ticket
    
    utils_log.log_url(log, "service url is:\t", service_url)
      
    return service_url

    



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

# SI unit prefixes
SI_K, SI_M, SI_G, SI_T = 10 ** 3, 10 ** 6, 10 ** 9, 10 ** 12

def convert_bytes(n):
    """Converts the given bytes into a string with the most appropriate
    unit power.
    
    Note that prefixes like M, G, T are power of 10 (ISO/IEC 80000-13:2008) and
    not power of 2."""        
    if   n >= SI_T:
        return '%.1f TB' % (float(n) / SI_T)
    elif n >= SI_G:
        return '%.1f GB' % (float(n) / SI_G)
    elif n >= SI_M:
        return '%.1f MB' % (float(n) / SI_M)
    elif n >= SI_K:
        return '%.1f kB' % (float(n) / SI_K)
    else:
        return '%d B' % n



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
#
#  http://cls-motu.sourceforge.net/
#
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
#
#
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


import sys
from cgi import log
if sys.version_info > (3, 0):
    import urllib
    import configparser as ConfigParser
else:
    import urllib2 as urllib
    import ConfigParser

import traceback
import platform
import sys
import os
import datetime
import logging
import logging.config

import argparse
from motu_utils import utils_configpath

# error code to use when exiting after exception catch
ERROR_CODE_EXIT=1

# the config file to load from
CFG_FILE = '~/motuclient/motuclient-python.ini'
LOG_CFG_FILE = './motu_utils/cfg/log.ini'

SECTION = 'Main'

# shared logger
log = None

# shared variables to download
_variables = []

# project libraries path
LIBRARIES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'motu_utils')
# Manage imports of project libraries
if not os.path.exists(LIBRARIES_PATH):
    sys.stderr.write('\nERROR: can not find project libraries path: %s\n\n' % os.path.abspath(LIBRARIES_PATH))
    sys.exit(1)
sys.path.append(LIBRARIES_PATH)

# Import project libraries
import utils_log
import utils_messages
import motu_api

def get_client_version():
    """Return the version (as a string) of this client.

    The value is automatically set by the maven processing build, so don't
    touch it unless you know what you are doing."""
    return motu_api.get_client_version()

def get_client_artefact():
    """Return the artifact identifier (as a string) of this client.

    The value is automatically set by the maven processing build, so don't
    touch it unless you know what you are doing."""
    return motu_api.get_client_artefact()

def load_options():
    """load options to handle"""

    _options = None

    # create option parser
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version=get_client_artefact() + ' v' + get_client_version())
    # add available options
    parser.add_argument( '--quiet', '-q',
                       help = "prevent any output in stdout",
                       action = 'store_const',
                       const = logging.WARN,
                       dest='log_level')

    parser.add_argument( '--verbose',
                       help = "print information in stdout",
                       action='store_const',
                       const = logging.DEBUG,
                       dest='log_level')

    parser.add_argument( '--noisy',
                       help = "print more information (traces) in stdout",
                       action='store_const',
                       const = utils_log.TRACE_LEVEL,
                       dest='log_level')

    parser.add_argument( '--user', '-u',
                       help = "the user name (string)")

    parser.add_argument( '--pwd', '-p',
                       help = "the user password (string)")

    parser.add_argument( '--auth-mode',
                       default = motu_api.AUTHENTICATION_MODE_CAS,
                       help = "the authentication mode: '" + motu_api.AUTHENTICATION_MODE_NONE  +
                              "' (for no authentication), '"+ motu_api.AUTHENTICATION_MODE_BASIC +
                              "' (for basic authentication), or '"+motu_api.AUTHENTICATION_MODE_CAS +
                              "' (for Central Authentication Service) [default: " + motu_api.AUTHENTICATION_MODE_CAS + "]")

    parser.add_argument( '--proxy-server',
                       help = "the proxy server (url)")

    parser.add_argument( '--proxy-user',
                       help = "the proxy user (string)")

    parser.add_argument( '--proxy-pwd',
                       help = "the proxy password (string)")

    parser.add_argument( '--motu', '-m',
                       help = "the motu server to use (url)")

    parser.add_argument( '--service-id', '-s',
                       help = "The service identifier (string)")

    parser.add_argument( '--product-id', '-d',
                       help = "The product (data set) to download (string)")

    parser.add_argument( '--date-min', '-t',
                       help = "The min date with optional hour resolution (string following format YYYY-MM-DD [HH:MM:SS])")

    parser.add_argument( '--date-max', '-T',
                       help = "The max date with optional hour resolution (string following format YYYY-MM-DD [HH:MM:SS])",
                       default = datetime.date.today().isoformat())

    parser.add_argument( '--latitude-min', '-y',
                       help = "The min latitude (float in the interval [-90 ; 90])")

    parser.add_argument( '--latitude-max', '-Y',
                       help = "The max latitude (float in the interval [-90 ; 90])")

    parser.add_argument( '--longitude-min', '-x',
                       help = "The min longitude (float)")

    parser.add_argument( '--longitude-max', '-X',
                       help = "The max longitude (float)")

    parser.add_argument( '--depth-min', '-z',
                       type = str,
                       help = "The min depth (float in the interval [0 ; 2e31] or string 'Surface')")

    parser.add_argument( '--depth-max', '-Z',
                       type = str,
                       help = "The max depth (float in the interval [0 ; 2e31] or string 'Surface')")

    parser.add_argument( '--variable', '-v',
                       help = "The variable (list of strings)",
                       dest="variable",
                       action='append',
                       nargs='+',
                       type=str)

    parser.add_argument( '--sync-mode', '-S',
                       help = "Sets the download mode to synchronous (not recommended)",
                       action='store_true',
                       dest='sync')

    parser.add_argument( '--describe-product', '-D',
                       help = "Get all updated information on a dataset. Output is in XML format",
                       action='store_true',
                       dest='describe')

    parser.add_argument( '--size',
                       help = "Get the size of an extraction. Output is in XML format",
                       action='store_true',
                       dest='size')

    parser.add_argument( '--out-dir', '-o',
                       help = "The output dir where result (download file) is written (string). If it starts with 'console', behaviour is the same as with --console-mode. ",
                       default=".")

    parser.add_argument( '--out-name', '-f',
                       help = "The output file name (string)",
                       default="data.nc")

    parser.add_argument( '--block-size',
                       type = int,
                       help = "The block used to download file (integer expressing bytes)",
                       default="65536")

    parser.add_argument( '--socket-timeout',
                       type = float,
                       help = "Set a timeout on blocking socket operations (float expressing seconds)")
    parser.add_argument( '--user-agent',
                       help = "Set the identification string (user-agent) for HTTP requests. By default this value is 'Python-urllib/x.x' (where x.x is the version of the python interpreter)")

    parser.add_argument( '--outputWritten',
                       help = "Optional parameter used to set the format of the file returned by the download request, only netcdf is supported. If not set, netcdf is used.")

    parser.add_argument( '--console-mode',
                       help = "Optional parameter used to display result on stdout, either URL path to download extraction file, or the XML content of getSize or describeProduct requests.",
                       action='store_true',
                       dest='console_mode')

    parser.add_argument( '--config-file',
                       help = "Path of the optional configuration file [default: %s]" % CFG_FILE,
                       action='append',
                       dest="config_file",
                       type=str)

    # create config parser
    conf_parser = ConfigParser.ConfigParser()

    _options = parser.parse_args()
    # flatten the variable lists of lists
    if not _options.variable is None:
        _options.variable = [var for varlists in _options.variable for var in varlists]
    
    # read configuration file name from cli arguments or use default
    # cant set default in parser.add_option due to optparse/argparse bug:
    # https://bugs.python.org/issue16399
    config_file = _options.config_file
    if config_file is None:
        config_file = [CFG_FILE]
    config_file=[os.path.expanduser(x) for x in config_file]
    conf_parser.read(config_file)

    # set default values by picking from the configuration file
    default_values = {}
    default_variables = []
    
    if (conf_parser.has_section(SECTION)):
        parser_options = conf_parser.options(SECTION)
        for option in conf_parser.options(SECTION):
            if hasattr(_options, option):
                if option == "variable":
                    variablesInCfgFile = conf_parser.get(SECTION, option)
                    if (not variablesInCfgFile is None) and variablesInCfgFile.strip():
                        allVariablesArray = variablesInCfgFile.split(",")
                        default_variables = default_variables + allVariablesArray
                        default_values[option] = default_variables
                else:
                    default_values[option] = conf_parser.get(SECTION, option)
    
        parser.set_defaults( **default_values )
        return parser.parse_args()
    else:
        return _options

def initLogger():
    logging.addLevelName(utils_log.TRACE_LEVEL, 'TRACE')
    logging.config.fileConfig(  os.path.join(os.path.dirname(__file__),LOG_CFG_FILE) )
    global log
    log = logging.getLogger(__name__)

    logging.getLogger().setLevel(logging.INFO)
#===============================================================================
# The Main function
#===============================================================================
def main():
    start_time = datetime.datetime.now()

    initLogger()

    try:
        # we prepare options we want
        _options = load_options()

        if _options.log_level != None:
            logging.getLogger().setLevel(int(_options.log_level))

        motu_api.execute_request(_options)
    except Exception as e:
        log.error( "Execution failed: %s", e )
        if hasattr(e, 'reason'):
          log.info( ' . reason: %s', e.reason )
        if hasattr(e, 'code'):
          log.info( ' . code  %s: ', e.code )
        if hasattr(e, 'read'):
          try:
            log.log( utils_log.TRACE_LEVEL, ' . detail:\n%s', e.read() )
          except:
            pass

        log.debug( '-'*60 )
        log.debug( "Stack trace exception is detailed hereafter:" )
        exc_type, exc_value, exc_tb = sys.exc_info()
        x = traceback.format_exception(exc_type, exc_value, exc_tb)
        for stack in x:
            log.debug( ' . %s', stack.replace('\n', '') )
        log.debug( '-'*60 )
        log.log( utils_log.TRACE_LEVEL, 'System info is provided hereafter:' )
        system, node, release, version, machine, processor = platform.uname()
        log.log( utils_log.TRACE_LEVEL, ' . system   : %s', system )
        log.log( utils_log.TRACE_LEVEL, ' . node     : %s', node )
        log.log( utils_log.TRACE_LEVEL, ' . release  : %s', release )
        log.log( utils_log.TRACE_LEVEL, ' . version  : %s', version )
        log.log( utils_log.TRACE_LEVEL, ' . machine  : %s', machine )
        log.log( utils_log.TRACE_LEVEL, ' . processor: %s', processor )
        log.log( utils_log.TRACE_LEVEL, ' . python   : %s', sys.version )
        log.log( utils_log.TRACE_LEVEL, ' . client   : %s', get_client_version() )
        log.log( utils_log.TRACE_LEVEL, '-'*60 )

        sys.exit(ERROR_CODE_EXIT)

    finally:
        log.debug( "Elapsed time : %s", str(datetime.datetime.now() - start_time) )

if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


import sys
if sys.version_info > (3, 0):
    import urllib.request, urllib.error
    from urllib.request import BaseHandler
    from urllib.parse import unquote
else:

    from urllib2 import BaseHandler
    from urllib2 import unquote



import logging

# trace level
TRACE_LEVEL = 1

def log_url(log, message, url, level = logging.DEBUG ):
    """Nicely logs the given url.
    
    Print out the url with the first part (protocol, host, port, authority,
    user info, path, ref) and in sequence all the query parameters.
    
    log: the log into which write the message
    message: a message to print before the url
    url: the url to log
    level: (optional) the log level to use"""
    
    urls = url.split('?')
    log.log( level, message + unquote(urls[0]) )
    if len(urls) > 1:
        for a in sorted(urls[1].split('&')):
            param = a.split('=')
            if( len(param) < 2 ):
              param.append('')
            log.log( level, ' . %s = %s', unquote(param[0]), unquote(param[1]) )

            
class HTTPDebugProcessor(BaseHandler):
    """ Track HTTP requests and responses with this custom handler.
    """
    def __init__(self, log, log_level=TRACE_LEVEL):
        self.log_level = log_level
        self.log = log

    def http_request(self, request):
        host, full_url = request.host, request.get_full_url()
        url_path = full_url[full_url.find(host) + len(host):]
        log_url ( self.log, "Requesting: ", request.get_full_url(), TRACE_LEVEL )
        self.log.log(self.log_level, "%s %s" % (request.get_method(), url_path))

        for header in request.header_items():
            self.log.log(self.log_level, " . %s: %s" % header[:])

        return request

    def http_response(self, request, response):
        code, msg, headers = response.code, response.msg, response.info().items()
        self.log.log(self.log_level, "Response:")
        self.log.log(self.log_level," HTTP/1.x %s %s" % (code, msg))
        
        for key, value in headers:
            self.log.log(self.log_level, " . %s: %s" % (key, value))

        return response            



from xml.etree import ElementTree
import os

def getPOMVersion():
    version="Unknown"
    try:
        # For production tree, while run from cur folder
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        version = __getPOMVersion(os.path.join(dname, "..", "pom.xml") )
    except:
        # For development tree
        try:
            version = __getPOMVersion(os.path.join(dname, "..", "..", "pom.xml") )
        except :
            version = __getPOMVersion(os.path.join(dname, "..", "..", "..", "pom.xml") )
    finally:
        return version
    
def __getPOMVersion(POM_FILE):
    namespaces = {'xmlns' : 'http://maven.apache.org/POM/4.0.0'}
    tree = ElementTree.parse(POM_FILE)
    root = tree.getroot()
    version = root.find("xmlns:version", namespaces=namespaces)
    return str(version.text)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

from threading import Thread, local

import time
import threading

# global stats
tsl = local()

class StopWatch(object):    
    TIME = "time"
    GLOBAL = "global"
    
    def __init__(self):
        # contains the computed times
        self.times = {}
        # contains the current timers
        self.timers = {}        
    
    def clear(self):
        self.timers = {}
        self.times = {}
        
    def start(self,label = GLOBAL):
        """Starts a new counter
           Returns the time the counter has been recorded.
        """
        self.timers[label] = self.__time()
        return self.timers[label]
        
    def stop(self,label=GLOBAL):
        """Stops the clock for the given counter.
        Returns the time at which the instance was stopped.
        """        
        self.times[label] = self.elapsed(label)
        del self.timers[label]
        
        return self.times[label]

    def isRunning(self, label=GLOBAL):
        return label in self.timers
        
    def elapsed(self,label=GLOBAL):
        """The number of seconds since the current time that the StopWatch
        object was created.  If stop() was called, it is the number
        of seconds from the instance creation until stop() was called.
        """
        t0 = self.times[label] if label in self.times else 0.
        t1 = self.timers[label] if label in self.timers else 0.
        t2 = self.__time() if label in self.timers else 0.
        
        return t0 + (t2 - t1)
        
    def getTimes(self):
        return self.times
          
    def __time(self):
        """Wrapper for time.time() to allow unit testing.
        """
        return time.time()
    
    def __str__(self):
        """Nicely format the elapsed time
        """
        keys = list(self.times.keys()) + [x for x in list(self.timers.keys()) if x not in list(self.times.keys())]
        txt = ""
        for key in keys:
          txt = txt + key + " : " + str(self.elapsed(key)) + " s " + ("(running)" if self.isRunning(key) else "(stopped)")+"\n"
        return txt

def localThreadStopWatch():    
    if not hasattr(tsl,'timer'):
        lock = threading.Lock()
        lock.acquire()
        try:        
            if not hasattr(tsl,'timer'):
                tsl.timer = StopWatch()
        finally:
            lock.release()
    return tsl.timer



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

import sys


if sys.version_info > (3, 0):
    import urllib.request, urllib.error
    from urllib.request import HTTPCookieProcessor, Request
    from urllib.request import HTTPSHandler, HTTPHandler, install_opener, build_opener, \
        HTTPRedirectHandler, ProxyHandler, HTTPBasicAuthHandler, HTTPPasswordMgrWithDefaultRealm
    from urllib.request import HTTPErrorProcessor as HTTPErrorProcessor_
    from http.client import HTTPConnection, HTTPSConnection
    from http.cookiejar import CookieJar
else:
    from httplib import HTTPConnection, HTTPSConnection
    from urllib2 import HTTPSHandler, install_opener, build_opener, \
    HTTPRedirectHandler, HTTPRedirectHandler, ProxyHandler, HTTPBasicAuthHandler, \
    HTTPPasswordMgrWithDefaultRealm, HTTPCookieProcessor, HTTPSHandler, HTTPHandler, Request
    from urllib2 import HTTPErrorProcessor as HTTPErrorProcessor_
    from cookielib import CookieJar

from motu_utils import utils_log
import logging
import ssl
import socket


class TLS1v2Connection(HTTPSConnection):
    """Like HTTPSConnection but more specific"""
    def __init__(self, host, **kwargs):
        HTTPSConnection.__init__(self, host, **kwargs)
 
    def connect(self):
        """Overrides HTTPSConnection.connect to specify TLS version"""
        # Standard implementation from HTTPSConnection, which is not
        # designed for extension, unfortunately
        sock = socket.create_connection((self.host, self.port),
                self.timeout, self.source_address)
        if getattr(self, '_tunnel_host', None):
            self.sock = sock
            self._tunnel()
 
        # This is the only difference; default wrap_socket uses SSLv23
        self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file,
                ssl_version=ssl.PROTOCOL_TLSv1_2)
 
class TLS1v2Handler(HTTPSHandler):
    """Like HTTPSHandler but more specific"""
    def __init__(self):
        HTTPSHandler.__init__(self)
 
    def https_open(self, req):
        return self.do_open(TLS1v2Connection, req)

# Overide default handler
install_opener(build_opener(TLS1v2Handler()))


class HTTPErrorProcessor(HTTPErrorProcessor_):
    def https_response(self, request, response):
        # Consider error codes that are not 2xx (201 is an acceptable response)
        code, msg, hdrs = response.code, response.msg, response.info()
        if code >= 300:
            response = self.parent.error('http', request, response, code, msg, hdrs)
        return response
    

class SmartRedirectHandler(HTTPRedirectHandler):

    def http_error_302(self, req, fp, code, msg, headers):
        result = HTTPRedirectHandler.http_error_302(
            self, req, fp, code, msg, headers)              
        result.status = code                                
        return result
    
    
def open_url(url, **kargsParam):
    """open an url and return an handler on it.
       arguments can be :
         headers : http headers to send
            headers = {"Accept": "text/plain", 
                       "User-Agent": "a user agent"
                      }
                      
         proxy : the proxy to use when connecting to the url
            proxy = { "url": "http://aproxy.server",
                      "port": 8080,
                      "user": "username",
                      "password": "userpassword"
                    }
          
         authentication: the authentication information
            authentication = { "mode": "basic",
                               "user": "username",
                               "password": "password" }
    """   
    data = None
    log = logging.getLogger("utils_http:open_url")
    kargs = kargsParam.copy()
    # common handlers
    handlers = [SmartRedirectHandler(),
                HTTPCookieProcessor(CookieJar()),
                HTTPHandler(),
                TLS1v2Handler(),
                utils_log.HTTPDebugProcessor(log),
                HTTPErrorProcessor()
                ]

    # add handlers for managing proxy credentials if necessary        
    if 'proxy' in kargs:
        urlProxy = ''
        if 'user' in kargs['proxy']:
            urlProxy = kargs['proxy']['user'] + ':' + kargs['proxy']['password'] + '@'

        urlProxy += kargs['proxy']['netloc'] 

        handlers.append( ProxyHandler({'http':urlProxy, 'https':urlProxy}) )
        handlers.append( HTTPBasicAuthHandler() )
        
        del kargs['proxy']
        
    if 'authentication' in kargs:
        # create the password manager
        #password_mgr = HTTPPasswordMgrWithDefaultRealm(ssl_version=ssl.PROTOCOL_TLSv1)
        password_mgr = HTTPPasswordMgrWithDefaultRealm()
        urlPart = url.partition('?')
        password_mgr.add_password(None, urlPart, kargs['authentication']['user'], kargs['authentication']['password'])
        # add the basic authentication handler
        handlers.append(HTTPSHandler(password_mgr))
        del kargs['authentication']
    
    if 'data' in kargs:
        data = kargs['data']
        del kargs['data']
    
    _opener = build_opener(*handlers)
    log.log(utils_log.TRACE_LEVEL, 'list of handlers:')
    for h in _opener.handlers:
        log.log(utils_log.TRACE_LEVEL, ' . %s', str(h))

    # create the request
    if( data != None ):
        r = Request(url, data, **kargs)
    else:
        r = Request(url, **kargs)

    # open the url, but let the exception propagates to the caller  
    return _opener.open(r)

def encode(options):    
    opts = []
    for k, vset in options.dict().items():
        for v in vset:
           opts.append('%s=%s' % (str(k), str(v).replace('#','%23').replace(' ','%20')))
    return '&'.join(opts)



"""  
The multimap data structure associates multiple values to a key. 

In this module the multimap is implemented by a dictionary in which each key is  
associated to a container, which can be a list, a dictionary, or a set. 

These containers are created, accessed, and extended using the usual array 
notation, e.g. m["a"] = 1, m["a"] = 2 creates a container for key "a" 
containing 1 and 2. An item within a container can be removed using the 
"remove"  method.

Requires Python 2.7.  
"""

import collections

class Map(object):
    """ Map wraps a dictionary. It is essentially an abstract class from which
    specific multimaps are subclassed. """
    def __init__(self):
        self._dict = {}
        
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._dict))
    
    __str__ = __repr__
        
    def __getitem__(self, key):
        return self._dict[key]
    
    def __setitem__(self, key, value):
        self._dict[key] = value
    
    def __delitem__(self, key):
        del self._dict[key]
    
    def insert(self, **values):
        for k, v in values.items():
           self.__setitem__(k, v)
    
    def remove(self, key, value):
        del self._dict[key]
    
    def dict(self):
        """ Allows access to internal dictionary, if necessary. Caution: multimaps 
        will break if keys are not associated with proper container."""
        return self._dict

class ListMultimap(Map):
    """ ListMultimap is based on lists and allows multiple instances of same value. """
    def __init__(self, **values):
        self._dict = collections.defaultdict(list)
        self.insert(**values)
        
    def __setitem__(self, key, value):
        self._dict[key].append(value)
    
    def remove(self, key, value):
        self._dict[key].remove(value)

class SetMultimap(Map):
    """ SetMultimap is based on sets and prevents multiple instances of same value. """
    def __init__(self):
        self._dict = collections.defaultdict(set)
        
    def __setitem__(self, key, value):
        self._dict[key].add(value)
    
    def remove(self, key, value):
        self._dict[key].remove(value)

class DictMultimap(Map):
    """ DictMultimap is based on dicts and allows fast tests for membership. """
    def __init__(self):
        self._dict = collections.defaultdict(dict)
        
    def __setitem__(self, key, value):
        self._dict[key][value] = True
    
    def remove(self, key, value):
        del self._dict[key][value]

def test():
    def test_multimap(m):
        print("__________________________________")
        print(m["a"])
        m["a"] = 1
        m["a"] = 2
        m["a"] = 2
        m["a"] = 3
        m["a"] = 4
        m.insert( b = 'v1', c = 'v2' )
        print(m)
        m.remove("a", 4)
        print(m)
        print(("a" in m.dict()))
        print(m["a"])
        m["a"] = 5
        m["b"] = 6
        print(m)
        del m["b"]
        print(m)
        print((3 in m["a"]))
        
    test_multimap(ListMultimap())
    test_multimap(SetMultimap())
    test_multimap(DictMultimap())

if __name__ == "__main__":
    test()



import os
import sys

from setuptools import setup, find_packages

# project libraries path
LIBRARIES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'src', 'python')

# Manage imports of project libraries
if not os.path.exists(LIBRARIES_PATH):
    sys.stderr.write('\nERROR: can not find project libraries path: %s\n\n' % os.path.abspath(LIBRARIES_PATH))
    sys.exit(1)
sys.path.append(LIBRARIES_PATH)

from motu_utils import pom_version

setup(
    name='motuclient',
    version=pom_version.getPOMVersion(),
    python_requires='>=2.7',
    description='Extract and download gridded data through a python command line from Motu web server. Used in CMEMS context http://marine.copernicus.eu/',
    long_description='Motu is a high efficient and robust Web Server which fills the gap between heterogeneous data providers to end users. Motu handles, extracts and transforms oceanographic huge volumes of data without performance collapse. This client enables to extract and download data through a python command line.',
    keywords=[
        'Copernicus',
        'CMEMS',
        'CLS',
        'Motu',
        'motuclient-python',
        'Dissemination Unit'
    ],

    author='Sylvain MARTY, CLS',
    author_email='smarty@cls.fr',
    platforms=['any'],

    url='https://github.com/clstoulouse/motu-client-python',
    license='LGPL',

    package_dir = {'': 'src/python', 'motu_utils': 'src/python/motu_utils'},
    packages=['motu_utils'],
    py_modules = ['motuclient', 'motu-client'],
    include_package_data=True,

    download_url='https://github.com/clstoulouse/motu-client-python/releases/',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Topic :: Scientific/Engineering :: GIS',
        'Environment :: Console',
        'Natural Language :: English',
        'Operating System :: OS Independent'
    ],

    entry_points={
        'console_scripts': [
            'motuclient = motuclient:main'
        ]
    }
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

import sys


if sys.version_info > (3, 0):
    import urllib.request, urllib.error
    from urllib.request import BaseHandler
    from urllib.parse import unquote, quote_plus, urlparse
else:
    from urllib import quote_plus, quote
    from urlparse import urlparse


import os
import re
import datetime
import time
import socket
from xml.dom import minidom
from pkg_resources import get_distribution

# Import project libraries
from motu_utils import utils_http, utils_stream, utils_cas, utils_log, utils_messages, utils_unit, \
    stop_watch, pom_version
from motu_utils import utils_collection
import logging

# constant for authentication modes
AUTHENTICATION_MODE_NONE  = 'none'
AUTHENTICATION_MODE_BASIC = 'basic'
AUTHENTICATION_MODE_CAS   = 'cas'

# constant for date time string format
DATETIME_FORMAT = "%Y-%m-%d% %H:%M:%S"


# shared logger
log = None

def get_client_version():
    """Return the version (as a string) of this client.
    
    The value is automatically set by the maven processing build, so don't 
    touch it unless you know what you are doing."""
    version = 'unknown'
    try:
        version = get_distribution('motuclient').version    
    except:
        version = pom_version.getPOMVersion()
    return version

def get_client_artefact():
    """Return the artifact identifier (as a string) of this client.
    
    The value is automatically set by the maven processing build, so don't 
    touch it unless you know what you are doing."""
    return 'motuclient-python'
    
def build_params(_options):
    """Function that builds the query string for Motu according to the given options"""
    temporal = ''
    geographic = ''
    vertical = ''
    other_opt = ''
    
    """
    Build the main url to connect to
    """
    query_options = utils_collection.ListMultimap()
    
    # describeProduct in XML format (sync) / productDownload (sync/async)
    if _options.describe:
        query_options.insert( action  = 'describeProduct',                
                              service = _options.service_id,
                              product = _options.product_id 
                            )   
    elif _options.size:   
        query_options.insert( action  = 'getSize',                
                              service = _options.service_id,
                              product = _options.product_id 
                            )        
    else:                        
        # synchronous/asynchronous mode
        if _options.sync:
            log.info('Synchronous mode set')
            query_options.insert( action  = 'productdownload',
                                  scriptVersion = quote_plus(get_client_version()),
                                  mode = 'console',
                                  service = _options.service_id,
                                  product = _options.product_id 
                                )                        
        else:
            log.info('Asynchronous mode set')
            query_options.insert( action  = 'productdownload',
                                  scriptVersion = quote_plus(get_client_version()),
                                  mode    = 'status',
                                  service = _options.service_id,
                                  product = _options.product_id 
                                )
    
    # geographical parameters
    if _options.extraction_geographic:
        query_options.insert( x_lo = _options.longitude_min,
                              x_hi = _options.longitude_max,
                              y_lo = _options.latitude_min,
                              y_hi = _options.latitude_max
                            )
    
    if _options.extraction_vertical:
        query_options.insert( z_lo = _options.depth_min,
                              z_hi = _options.depth_max
                            )
        
    """ MOTU-172
    if _options.extraction_output:
        query_options.insert(output=_options.outputWritten)
    else:"""
    query_options.insert(output="netcdf")
    
    if _options.extraction_temporal:
        # date are strings, and they are send to Motu "as is". If not, we convert them into string
        if _options.date_min is not None or _options.date_min != None:            
            date_min = _options.date_min
            if not isinstance(date_min, str):
               date_min = date_min.strftime(DATETIME_FORMAT)
            query_options.insert( t_lo = date_min )
            
        if _options.date_max is not None or _options.date_max != None:            
            date_max = _options.date_max
            if not isinstance(date_max, str):
               date_max = date_max.strftime(DATETIME_FORMAT)
            query_options.insert( t_hi = date_max )

    variable = _options.variable
    if variable is not None:
        for i, opt in enumerate(variable):
            query_options.insert( variable = opt )
    
    
    if _options.console_mode or _options.out_dir.startswith("console") :
        query_options.insert(mode="console")
        
    return utils_http.encode(query_options)

def check_options(_options):    
    """function that checks the given options for coherency."""    
    
    # Check Mandatory Options
    if (_options.auth_mode != AUTHENTICATION_MODE_NONE and 
        _options.auth_mode != AUTHENTICATION_MODE_BASIC and
        _options.auth_mode != AUTHENTICATION_MODE_CAS):
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.invalid'] % (_options.auth_mode, 'auth-mode', [AUTHENTICATION_MODE_NONE, AUTHENTICATION_MODE_BASIC, AUTHENTICATION_MODE_CAS]))
       
    # if authentication mode is set we check both user & password presence
    if (_options.user == None and
        _options.auth_mode != AUTHENTICATION_MODE_NONE):
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory.user'] % ('user', _options.auth_mode))

    # check that if a user is set, a password should be set also
    if (_options.pwd == None and
        _options.user != None):
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory.password'] % ('pwd', _options.user))
    
    #check that if a user is set, an authentication mode should also be set
    if (_options.user != None and
        _options.auth_mode == AUTHENTICATION_MODE_NONE):
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory.mode'] % (AUTHENTICATION_MODE_NONE, 'auth-mode', _options.user))
    
    # those following parameters are required
    if _options.motu == None :
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory'] % 'motu')
    
    if _options.service_id == None :
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory'] % 'service-id')
    
    if _options.product_id == None :
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory'] % 'product-id')
    
    if _options.out_dir == None :
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.mandatory'] % 'out-dir')
    
    out_dir = _options.out_dir
    if not out_dir.startswith("console"):
        # check directory existence
        if not os.path.exists(out_dir):
            raise Exception(
                utils_messages.get_external_messages()['motuclient.exception.option.outdir-notexist'] % out_dir)
        # check whether directory is writable or not
        if not os.access(out_dir, os.W_OK):
            raise Exception(
                utils_messages.get_external_messages()['motuclient.exception.option.outdir-notwritable'] % out_dir)
    
        if _options.out_name == None :
            raise Exception(
                utils_messages.get_external_messages()['motuclient.exception.option.mandatory'] % 'out-name')

    # Check PROXY Options
    _options.proxy = False
    if (_options.proxy_server != None) and (len(_options.proxy_server) != 0):
        _options.proxy = True
        # check that proxy server is a valid url
        url = _options.proxy_server
        p = re.compile('^(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%@!\-\/]))?')
        m = p.match(url)
        
        if not m :
            raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.not-url'] % ('proxy-server', url))
        # check that if proxy-user is defined then proxy-pwd shall be also, and reciprocally.
        if (_options.proxy_user != None) != ( _options.proxy_pwd != None ) :
            raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.linked'] % ('proxy-user', 'proxy-name'))
    
        
    # Check VERTICAL Options
    _options.extraction_vertical = False
    if _options.depth_min != None or _options.depth_max != None :
        _options.extraction_vertical = True
        
    # Check TEMPORAL  Options
    _options.extraction_temporal = False
    if _options.date_min != None or _options.date_max != None :
         _options.extraction_temporal = True
    
    """ MOTU-172
    #Check OUTPUT Options
    _options.extraction_output = False
    if _options.outputWritten != None :
        _options.extraction_output = True
    """
    # Check GEOGRAPHIC Options
    _options.extraction_geographic = False
    if _options.latitude_min != None or _options.latitude_max != None or _options.longitude_min != None or _options.longitude_max != None :
        _options.extraction_geographic = True
        
        check_latitude(_options.latitude_min, 'latitude_min')
        check_latitude(_options.latitude_max, 'latitude_max')
        
        check_coordinate(_options.longitude_min, 'longitude_min')
        check_coordinate(_options.longitude_max, 'longitude_max')

def check_coordinate(coord, msg):
    if( coord == None ):
        raise Exception(
            utils_messages.get_external_messages()['motuclient.exception.option.geographic-box'] % msg)
    try:
        return float(coord)
    except ValueError:
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.invalid'] % (coord, msg, 'floating point number'))

def check_latitude(lat, msg):
    tempvalue = check_coordinate(lat, msg)
    if tempvalue < -90 or tempvalue > 90 :
        raise Exception(utils_messages.get_external_messages()['motuclient.exception.option.out-of-range'] % (msg, str(tempvalue)))

def total_seconds(td):
    return total_milliseconds(td) / 10**3 

def total_milliseconds(td):
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**3 

def get_url_config(_options, data = None):
    # prepare arguments    
    kargs = {}
    # proxy
    if _options.proxy:
        #proxyUrl = _options.proxy_server.partition(':')
        proxyUrl = urlparse(_options.proxy_server)
        kargs['proxy'] = { "scheme": proxyUrl.scheme,
                           "netloc": proxyUrl.netloc }
        if _options.proxy_user != None:
            kargs['proxy']['user']     = _options.proxy_user
            kargs['proxy']['password'] = _options.proxy_pwd
    # authentication
    if _options.auth_mode == AUTHENTICATION_MODE_BASIC:
        kargs['authentication'] = { 'mode'    : 'basic',
                                    'user'    : _options.user,
                                    'password': _options.pwd }
    # headers
    kargs['headers'] = {"X-Client-Id"     : get_client_artefact(),
                        "X-Client-Version": quote_plus(get_client_version())}
    # data
    if data != None:
        kargs['data'] = data
    
    return kargs

def get_requestUrl(dl_url, server, _options, **options):
    """ Get the request url."""    
    stopWatch = stop_watch.localThreadStopWatch()    
    start_time = datetime.datetime.now()
    stopWatch.start('get_request')
    log.info( "Requesting file to download (this can take a while)..." ) 
    
    # Get request id        
    m = utils_http.open_url(dl_url, **options)
    responseStr = m.read()
    dom = minidom.parseString(responseStr)
    node = dom.getElementsByTagName('statusModeResponse')[0]
    status = node.getAttribute('status')
    if status == "2":
        msg = node.getAttribute('msg')
        log.error(msg)
        get_req_url = None
    else:
        requestId = node.getAttribute('requestId')
        # Get request url
        get_req_url = server + '?action=getreqstatus&requestid=' + requestId + "&service=" + _options.service_id + "&product=" + _options.product_id
        
    stopWatch.stop('get_request')
    
    return get_req_url
    
def wait_till_finished(reqUrlCAS, **options):    
    stopWatch = stop_watch.localThreadStopWatch()    
    start_time = datetime.datetime.now()

  
lastProgressPercentValue=0.0
  
def dl_2_file(dl_url, fh, block_size = 65535, isADownloadRequest = None, **options):
    """ Download the file with the main url (of Motu) file.
     
    Motu can return an error message in the response stream without setting an
    appropriate http error code. So, in that case, the content-type response is
    checked, and if it is text/plain, we consider this as an error.
    
    dl_url: the complete download url of Motu
    fh: file handler to use to write the downstream"""    
    
    stopWatch = stop_watch.localThreadStopWatch()    
    start_time = datetime.datetime.now()
    lastProgressPercentValue=0.0
    log.info("Downloading file (this can take a while)..." )    

    # download file
    temp = None
    if not fh.startswith("console"):
        temp = open(fh, 'w+b')
        
    try:
        stopWatch.start('processing')
        
        m = utils_http.open_url(dl_url, **options)
        try:
            # check the real url (after potential redirection) is not a CAS Url scheme
            match = re.search(utils_cas.CAS_URL_PATTERN, m.url)
            if match is not None:
                service, _, _ = dl_url.partition('?')
                redirection, _, _ = m.url.partition('?')
                raise Exception(
                    utils_messages.get_external_messages()['motuclient.exception.authentication.redirected'] % (service, redirection))

            # check that content type is not text/plain
            headers = m.info()
            if "Content-Type" in headers and len(headers['Content-Type']) > 0 and isADownloadRequest and (headers['Content-Type'].startswith('text') or headers['Content-Type'].find('html') != -1):
                raise Exception(utils_messages.get_external_messages()['motuclient.exception.motu.error'] % m.read())
              
            log.info( 'File type: %s' % headers['Content-Type'] )
            # check if a content length (size of the file) has been send
            size = -1
            if "Content-Length" in headers:        
                try:
                    # it should be an integer
                    size = int(headers["Content-Length"]) 
                    log.info( 'File size: %s (%i B)' % (utils_unit.convert_bytes(size), size))
                except Exception as e:
                    size = -1
                    log.warn( 'File size is not an integer: %s' % headers["Content-Length"] )                      
            elif temp is not None:
                log.warn( 'File size: %s' % 'unknown' )

            processing_time = datetime.datetime.now();        
            stopWatch.stop('processing')        
            stopWatch.start('downloading')
        
            # performs the download           
            log.info( 'Downloading file %s' % os.path.abspath(fh) )
        
            def progress_function(sizeRead):
                global lastProgressPercentValue
                percent = sizeRead*100./size
                if percent - lastProgressPercentValue> 1 or (lastProgressPercentValue != 100 and percent >= 100) :
                    log.info( "- %s (%.1f%%)", utils_unit.convert_bytes(size).rjust(8), percent)
                    lastProgressPercentValue = percent
                    
            def none_function(sizeRead):
                global lastProgressPercentValue
                percent = 100
                log.info( "- %s (%.1f%%)", utils_unit.convert_bytes(size).rjust(8), percent)
                lastProgressPercentValue = percent
            
            if temp is not None:
                read = utils_stream.copy(m, temp, progress_function if size != -1 else none_function, block_size)
            else:
                if isADownloadRequest:
                    #Console mode, only display the NC file URL on stdout
                    read = len(m.url)
                    print((m.url))
                else:
                    import io
                    output = io.StringIO()
                    utils_stream.copy(m, output, progress_function if size != -1 else none_function, block_size)
                    read = len(output.getvalue())
                    print((output.getvalue()))
                
            end_time = datetime.datetime.now()
            stopWatch.stop('downloading')
            
            log.info( "Processing  time : %s", str(processing_time - init_time) )
            log.info( "Downloading time : %s", str(end_time - processing_time) )
            log.info( "Total time       : %s", str(end_time - init_time) )
            log.info( "Download rate    : %s/s", utils_unit.convert_bytes((read / total_milliseconds(end_time - start_time)) * 10 ** 3))
        except Exception as e:
            log.error( "Download failed: %s", e )
            if hasattr(e, 'reason'):
              log.info( ' . reason: %s', e.reason )
            if hasattr(e, 'code'):
              log.info( ' . code  %s: ', e.code )
            if hasattr(e, 'read'):
              try:
                log.log( utils_log.TRACE_LEVEL, ' . detail:\n%s', e.read() )
              except:
                pass
    
            log.debug( '-'*60 )
            log.debug( "Stack trace exception is detailed herafter:" )
            exc_type, exc_value, exc_tb = sys.exc_info()
            x = traceback.format_exception(exc_type, exc_value, exc_tb)
            for stack in x:
                log.debug( ' . %s', stack.replace('\n', '') )
            log.debug( '-'*60 )
            log.log( utils_log.TRACE_LEVEL, 'System info is provided hereafter:' )
            system, node, release, version, machine, processor = platform.uname()
            log.log( utils_log.TRACE_LEVEL, ' . system   : %s', system )
            log.log( utils_log.TRACE_LEVEL, ' . node     : %s', node )
            log.log( utils_log.TRACE_LEVEL, ' . release  : %s', release )
            log.log( utils_log.TRACE_LEVEL, ' . version  : %s', version )
            log.log( utils_log.TRACE_LEVEL, ' . machine  : %s', machine )
            log.log( utils_log.TRACE_LEVEL, ' . processor: %s', processor )
            log.log( utils_log.TRACE_LEVEL, ' . python   : %s', sys.version )
            log.log( utils_log.TRACE_LEVEL, ' . client   : %s', get_client_version() )
            log.log( utils_log.TRACE_LEVEL, '-'*60 )
        
        finally:
            m.close()
    finally:
        if temp is not None:
            temp.flush()
            temp.close()

    # raise exception if actual size does not match content-length header
    if temp is not None and size >= 0 and read < size:
        raise Exception(
            utils_messages.get_external_messages()['motuclient.exception.download.too-short'] % (read, size))

def execute_request(_options):
    """
    the main function that submit a request to motu. Available options are:
    
    * Proxy configuration (with eventually user credentials)
      - proxy_server: 'http://my-proxy.site.com:8080'
      - proxy_user  : 'john'
      - proxy_pwd   :'doe'

    * Autorisation mode: 'cas', 'basic', 'none'
      - auth_mode: 'cas'
      
    * User credentials for authentication 'cas' or 'basic'
      - user: 'john'
      - pwd:  'doe'
    
    * Motu service URL
      - motu: 'http://atoll-dev.cls.fr:30080/mis-gateway-servlet/Motu'
    
    * Dataset identifier to download
      - product_id: 'dataset-duacs-global-nrt-madt-merged-h'
    
    * Service identifier to use for retrieving dataset
      - service_id: 'http://purl.org/myocean/ontology/service/database#yourduname'
    
    * Geographic extraction parameters
      - latitude_max :  10.0
      - latitude_min : -10.0
      - longitude_max: -0.333333333369
      - longitude_min:  0.0

    * Vertical extraction parameters
      - depth_max: 1000
      - depth_min: 0
    
    * Temporal extraction parameters, as a datetime instance or a string (format: '%Y-%m-%d %H:%M:%S')
      - date_max: 2010-04-25 12:05:36
      - date_min: 2010-04-25

    * Variable extraction
      - variable: ['variable1','variable2']
      
    * The file name and the directory of the downloaded dataset
      - out_dir : '.'
      - out_name: 'dataset'
      
    * The block size used to perform download
      - block_size: 12001
      
    * The socket timeout configuration
      - socket_timeout: 515

    * The user agent to use when performing http requests
      - user_agent: 'motu-api-client' 

    """
    global log
    global init_time

    init_time = datetime.datetime.now()
    stopWatch = stop_watch.localThreadStopWatch()
    stopWatch.start()
    try:
        log = logging.getLogger("motu_api")

        # at first, we check given options are ok
        check_options(_options)

        # print some trace info about the options set
        log.log(utils_log.TRACE_LEVEL, '-' * 60)

        for option in dir(_options):
            if not option.startswith('_'):
                log.log(utils_log.TRACE_LEVEL, "%s=%s" % (option, getattr(_options, option)))

        log.log(utils_log.TRACE_LEVEL, '-' * 60)

        # start of url to invoke
        url_service = _options.motu

        # parameters of the invoked service
        url_params  = build_params(_options)

        url_config = get_url_config(_options)

        # check if question mark is in the url
        questionMark = '?'
        if url_service.endswith(questionMark) :
            questionMark = ''
        url = url_service+questionMark+url_params

        if _options.describe == True or _options.size == True: 
            _options.out_name = _options.out_name.replace('.nc','.xml')

        # set-up the socket timeout if any
        if _options.socket_timeout != None:
            log.debug("Setting timeout %s" % _options.socket_timeout)
            socket.setdefaulttimeout(_options.socket_timeout)

        if _options.auth_mode == AUTHENTICATION_MODE_CAS:
            stopWatch.start('authentication')
            # perform authentication before acceding service
            download_url = utils_cas.authenticate_CAS_for_URL(url,
                                                              _options.user,
                                                              _options.pwd, **url_config)
            url_service =  download_url.split("?")[0]
            stopWatch.stop('authentication')
        else:
            # if none, we do nothing more, in basic, we let the url requester doing the job
            download_url = url

        # create a file for storing downloaded stream
        fh = os.path.join(_options.out_dir,_options.out_name)
        if _options.console_mode:
            fh = "console"
            
        try:
            # Synchronous mode
            if _options.sync == True or _options.describe == True or _options.size == True:
                is_a_download_request=False
                if _options.describe == False and _options.size == False:
                    is_a_download_request=True
                dl_2_file(download_url, fh, _options.block_size, is_a_download_request, **url_config)
                log.info( "Done" )
            # Asynchronous mode
            else:
                stopWatch.start('wait_request')
                requestUrl = get_requestUrl(download_url, url_service, _options, **url_config)    
                
                if requestUrl != None:    
                    # asynchronous mode
                    status = "0"
                    dwurl = ""
                    msg = ""
                
                    while True:
                        if _options.auth_mode == AUTHENTICATION_MODE_CAS:
                            stopWatch.start('authentication')
                            # perform authentication before acceding service
                            requestUrlCas = utils_cas.authenticate_CAS_for_URL(requestUrl,
                                                                               _options.user,
                                                                               _options.pwd, **url_config)
                            stopWatch.stop('authentication')
                        else:
                            # if none, we do nothing more, in basic, we let the url requester doing the job
                            requestUrlCas = requestUrl
                        
                        m = utils_http.open_url(requestUrlCas, **url_config)
                        motu_reply = m.read()
                        dom = None
                        
                        try:
                            dom = minidom.parseString(motu_reply)
                        except:
                            log.error(motu_reply)
                            dom = None
                        
                        if dom:    
                            for node in dom.getElementsByTagName('statusModeResponse'):
                                status = node.getAttribute('status')
                                dwurl = node.getAttribute('remoteUri')
                                msg = node.getAttribute('msg')
                        else:
                            status = "4"

                        # Check status
                        if status == "0" or status == "3": # in progress/pending
                            log.info('Product is not yet available (request in progress)')
                            time.sleep(10)
                        else: # finished (error|success)
                            break

                    if status == "2": 
                        log.error(msg) 
                    if status == "4":
                        log.error("Motu server API interaction appears to have failed, server response is invalid")
                    if status == "1": 
                        log.info('The product is ready for download')
                        if dwurl != "":
                            dl_2_file(dwurl, fh, _options.block_size, not (_options.describe or _options.size), **url_config)
                            log.info( "Done" )
                        else:
                            log.error("Couldn't retrieve file")
                    
                stopWatch.stop('wait_request')        
                            
        except:
            try:
                if (os.path.isfile(fh)):
                    os.remove(fh)
            except:
                pass
            raise
    finally:
        stopWatch.stop()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python motu client
#
# Motu, a high efficient, robust and Standard compliant Web Server for Geographic
#  Data Dissemination.
# 
#  http://cls-motu.sourceforge.net/
# 
#  (C) Copyright 2009-2010, by CLS (Collecte Localisation Satellites) -
#  http://www.cls.fr - and Contributors
# 
# 
#  This library is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
# 
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
#  License for more details.
# 
#  You should have received a copy of the GNU Lesser General Public License
#  along with this library; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

from motu_utils import utils_configpath
import os, sys
_messages = None

MESSAGES_FILE = utils_configpath.getConfigPath() + '/messages.properties'


def _process_content(content):
    propDict = dict()
    for propLine in content:
        propDef = propLine.strip()
        if len(propDef) == 0:
            continue
        if propDef[0] in ('!', '#'):
            continue
        punctuation = [propDef.find(c) for c in ':= '] + [len(propDef)]
        found = min([pos for pos in punctuation if pos != -1])
        name = propDef[:found].rstrip()
        value = propDef[found:].lstrip(":= ").rstrip()
        propDict[name] = value

    _messages = propDict
    return _messages


def _process_file_py3(path):
    with open(path) as propFile:
        return _process_content(propFile.readlines())


def _process_file_py2(path):
    propFile = file(path, "rU")
    processed_content = _process_content(propFile)
    propFile.close()
    return processed_content


def get_external_messages():
    """Return a table of externalized messages.

    The table is lazzy instancied (loaded once when called the first time)."""
    global _messages
    if _messages is None:
        if sys.version_info > (3, 0):
            return _process_file_py3(MESSAGES_FILE)
        return _process_file_py2(MESSAGES_FILE)